#!/usr/bin/python3

# Create gcode file for printing a box with walls in low warping mode
# Details on:
# https://homofaciens.de/technics-machines-3D-printer-Longer-LK5Pro-warping_en.htm

import os
import math

fileName = "Box-with-walls.gcode"

X = 0.0
Y = 0.0
Z = 0.0
E = 0.0

Xmax = 300.0
Ymax = 300.0
Zmax = 400.0

F_travel = 7800.0
F_1st_layer = 600
F_print = 1800
F_retract = 6000
Retract = 3.000
Bed_temp = 100
Nozzle_temp = 240

box_l = 100.000
box_w = 50.000
# Set box_h to 0.00 to print no walls
box_h = 0.000
box_wall = 2.400

E_width = 0.600
# Extrusion width 0.5mm, data from box-ABS.gcode, created with Slic3r
# In file line 586 to 591:
# Extruder goes from E665.97304 to E669.65044
# while X movement from X198.379 to X101.621
# Extruder movement per mm X movement per mm layer height (0.2mm was used in example file):
ExtrusionRate = (((669.65044 - 665.97304) / (198.379 - 101.621)) / 0.2) / 0.5
ExtrusionRate = ExtrusionRate * 1.08 # Filament factor


LAYER_H_1st = 0.350
LAYER_H = 0.200

# Height of bottom plate is set by layer number
# Heigt = LAYER_H_1st + (LAYERS - 1) * LAYER_H
LAYERS = 10


########################################################################
def driveToXYZ(driveX, driveY, driveZ, retract):
    global E

    # retract 0 = no retract
    # retract 1 = only retract
    # retract 2 = only forward
    # retract 3 = retract at beginning, forward at end of movement

    if retract == 1 or retract == 3:
      storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
      storeFile.write("G1 E" + format(-Retract, ".5f") + " F" + format(F_retract, ".3f") + "\n")
      storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
      E = 0
    storeFile.write("G1 F" + format(F_travel, ".3f") + "\n")
    storeFile.write("G1 Z" + format(driveZ, ".3f") + "\n")
    storeFile.write("G1 X" + format(driveX, ".3f") + " Y" + format(driveY, ".3f") + "\n")
    if retract == 2 or retract == 3:
      storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
      storeFile.write("G1 E" + format(Retract, ".5f") + " F" + format(F_retract, ".3f") + "\n")
      storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
      E = 0


########################################################################
def extrudeToXYZ(eXfrom, eYfrom, eZfrom, eXto, eYto, eZto, layer_h, E_width, E, F = 0.0, makeDots = False):
    global ExtrusionRate
    global dotDistance
    global currentDot

    travel_X = math.pow(eXfrom - eXto, 2)
    travel_Y = math.pow(eYfrom - eYto, 2)
    travel_Z = math.pow(eZfrom - eZto, 2)
    travel_l = math.sqrt(travel_X + travel_Y + travel_Z)

    if F != 0.0:
      storeFile.write("G1 F" + format(F, ".3f") + "\n")

    X1 = eXfrom
    Y1 = eYfrom
    Z1 = eZfrom

    if makeDots:
      ratioX = (eXto - eXfrom) / travel_l
      ratioY = (eYto - eYfrom) / travel_l
      ratioZ = (eZto - eZfrom) / travel_l
      while travel_l > currentDot:
        X2 = X1 + ratioX * currentDot
        Y2 = Y1 + ratioY * currentDot
        Z2 = Z1 + ratioZ * currentDot
        E = E + currentDot * ExtrusionRate * layer_h * E_width
        storeFile.write("G1 X" + format(X2, ".3f") + " Y" + format(Y2, ".3f") + " Z" + format(Z2, ".3f") + " E" + format(E, ".5f") + "\n")
        Ztop = Z2
        Z2 = Z2 - Z_HUB
        if Z2 < 0.1:
          Z2 = 0.1
        storeFile.write("G1 Z" + format(Z2, ".3f") + "\n")
        Z2 = Ztop
        storeFile.write("G1 Z" + format(Z2, ".3f") + "\n")
        travel_l = travel_l - currentDot
        currentDot = dotDistance
        X1 = X2
        Y1 = Y2
        Z1 = Z1
      if travel_l > 0.0:
        X2 = X1 + ratioX * travel_l
        Y2 = Y1 + ratioY * travel_l
        Z2 = Z1 + ratioZ * travel_l
        E = E + travel_l * ExtrusionRate * layer_h * E_width
        storeFile.write("G1 X" + format(X2, ".3f") + " Y" + format(Y2, ".3f") + " Z" + format(Z2, ".3f") + " E" + format(E, ".5f") + "\n")
        currentDot = currentDot - travel_l


    else:
      E = E + travel_l * ExtrusionRate * layer_h * E_width
      # print("travel_l = ", travel_l, ", E = ", E, ", Eold = ", Eold)
      storeFile.write("G1 X" + format(eXto, ".3f") + " Y" + format(eYto, ".3f") + " Z" + format(eZto, ".3f") + " E" + format(E, ".5f") + "\n")

    return E

########################################################################
def createPyramid(pX, pY, pZ, eX, eY, layerNo, Holow, Zsafe):

    global E

    if Zsafe > 0.0:
      driveToXYZ(pX, pY, pZ + layerNo * LAYER_H + Zsafe, 1)
      driveToXYZ(pX, pY, pZ, 2)
      E = 0
    if Zsafe < 0.0:
      driveToXYZ(pX, pY, pZ, 0)

    if Zsafe == 0.0:
      driveToXYZ(pX, pY, pZ + layerNo * LAYER_H + Zsafe, 0)
      driveToXYZ(pX, pY, pZ, 0)

    jStart = -1
    for Layer in range(layerNo):


      X = pX + Layer * E_width
      Y = pY + Layer * E_width
      Z = pZ + Layer * LAYER_H

      driveToXYZ(X, Y, Z, False)

      jStart = jStart + 1

      if Z == LAYER_H_1st: # Set print speed
        storeFile.write("G1 F" + format(F_1st_layer, ".3f") + "\n")
      else:
        storeFile.write("G1 F" + format(F_print, ".3f") + "\n")

      Xold = X
      Yold = Y

      if Z == LAYER_H_1st:
        layer_h = LAYER_H_1st
      else:
        layer_h = LAYER_H

      edge_min = eX
      if edge_min > eY:
        edge_min = eY

      if Holow:
        jEnd = jStart + 1
      else:
        jEnd = int((edge_min / E_width) / 2)

      for j in range(jStart, jEnd):
        # print("j=", j)
        X = X + (eX - j * E_width * 2 - E_width)
        E = extrudeToXYZ(Xold, Yold, Z, X, Y, Z, layer_h, E_width, E, 0)
        Xold = X
        Yold = Y

        Y = Y + (eY - j * E_width * 2 - E_width)
        E = extrudeToXYZ(Xold, Yold, Z, X, Y, Z, layer_h, E_width, E, 0)
        Xold = X
        Yold = Y

        X = X - (eX - j * E_width * 2 - E_width)
        E = extrudeToXYZ(Xold, Yold, Z, X, Y, Z, layer_h, E_width, E, 0)
        Xold = X
        Yold = Y

        Y = Y - (eY - j * E_width * 2 - E_width * 2)
        E = extrudeToXYZ(Xold, Yold, Z, X, Y, Z, layer_h, E_width, E, 0)
        X = X + E_width
        Xold = X
        Yold = Y


########################################################################
########################################################################
########################################################################

storeFile = open(fileName, "w")

storeFile.write("M190 S" + format(Bed_temp, ".3f") + " ; set bed temperature and wait for it to be reached" + "\n")
storeFile.write("M104 S" + format(Nozzle_temp, ".3f") + " ; set temperature" + "\n")
storeFile.write("G0 Z 10.000 ; Lift Z axis" + "\n")
storeFile.write("G28 ; home all axes" + "\n")
storeFile.write("G1 Z5 F5000 ; lift nozzle" + "\n")
storeFile.write("M109 S" + format(Nozzle_temp, ".3f") + " ; set temperature and wait for it to be reached" + "\n")
storeFile.write("G21 ; set units to millimeters" + "\n")
storeFile.write("G90 ; use absolute coordinates" + "\n")
storeFile.write("M82 ; use absolute distances for extrusion" + "\n")
storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
storeFile.write("G1 Z10.000 F" + format(F_travel, ".3f") + "\n")
storeFile.write("G1 E10.000 F120.000" + "\n")
storeFile.write("G92 E0" + "\n")
storeFile.write("G1 E" + format(-Retract, ".5f") + " F" + format(F_retract, ".3f") + "\n")
storeFile.write("G1 Z" + format(LAYER_H_1st, ".3f") + " F" + format(F_travel, ".3f") + "\n")
storeFile.write("G92 E0" + "\n")



E = 0
Xstart = (Xmax - box_l) / 2
Ystart = (Ymax - box_w) / 2


storeFile.write("G1 X" + format(Xstart, ".3f") + " Y" + format(Ystart, ".3f") + " F" + format(F_travel, ".3f") + "\n")
storeFile.write("G1 E" + format(Retract, ".5f") + " F" + format(F_retract, ".3f") + "\n")
storeFile.write("G92 E0" + "\n")
E = 0
storeFile.write("G1 F" + format(F_travel, ".3f") + "\n")

Z = LAYER_H_1st
# Extrude around maximum points
eXf = Xstart - 3.0
eYf = Ystart - 3.0
eXto = eXf
eYto = eYf + box_w + 6.0
driveToXYZ(eXf, eYf, Z, 3)
E = 0
E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, LAYER_H_1st, E_width, E, F_1st_layer, False)

eXf = eXto
eYf = eYto
eXto = eXf + box_l + 6.0
eYto = eYf
E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, LAYER_H_1st, E_width, E, F_1st_layer, False)

eXf = eXto
eYf = eYto
eXto = eXf
eYto = eYf - box_w - 6.0
E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, LAYER_H_1st, E_width, E, F_1st_layer, False)

eXf = eXto
eYf = eYto
eXto = Xstart - 3.0
eYto = eYf
E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, LAYER_H_1st, E_width, E, F_1st_layer, False)



# Create initial pentahedron

runs = 0
if box_l > box_w:
  edge_Y = LAYERS * 2 * E_width
  runs = (box_w - LAYERS * 2 * E_width) / (E_width * 2)
  edge_X = box_l - runs * 2 * E_width

else:
  edge_X = LAYERS * 2 * E_width
  runs = box_l / (E_width * 2) - LAYERS * 2
  edge_Y = box_w - runs * 2 * E_width

X = Xstart + runs * E_width
Y = Ystart + runs * E_width
Z = LAYER_H_1st
print("edge_X = ", edge_X, ", edge_Y = ", edge_Y)
driveToXYZ(X, Y, Z, 3)
E = 0
createPyramid(X, Y, Z, edge_X, edge_Y, LAYERS, False, -1.0)

# Grow holow pyramids
for j in range(int(runs)):
  X = X - E_width
  Y = Y - E_width
  edge_X = edge_X + 2 * E_width
  edge_Y = edge_Y + 2 * E_width
  createPyramid(X, Y, Z, edge_X, edge_Y, LAYERS, True, 0.01)


# Grow edges
for j in range(1, LAYERS):
  Z = Z + LAYER_H
  createPyramid(X, Y, Z, edge_X, edge_Y, LAYERS - j, True, 0.01)


if Z < box_h: # Grow wall socket
  wall_layers = int(box_wall / E_width)
  for j in range(wall_layers - 1):
    Z = LAYER_H_1st + LAYERS * LAYER_H
    X1 = X + (wall_layers - j -  1) * E_width
    Y1 = Y + (wall_layers - j -  1) * E_width
    EX = edge_X - (wall_layers - j -  1) * E_width * 2
    EY = edge_Y - (wall_layers - j -  1) * E_width * 2
    if Z < box_h:
      createPyramid(X1, Y1, Z, EX, EY, j + 1, True, 0.01)


# Grow walls
while Z < box_h:
  createPyramid(X, Y, Z, edge_X, edge_Y, wall_layers, True, 0.01)
  Z = Z + LAYER_H


# Final Gcode lines, turn off nozzle and bed et cetera...
storeFile.write("G92 E0" + "\n")
storeFile.write("M104 S0 ; turn off hotend temperature" + "\n")
storeFile.write("M140 S0 ; turn off bed temperature" + "\n")
storeFile.write("G28 X0  ; home X axis" + "\n")
storeFile.write("M84     ; disable motors" + "\n")

print("Gcode file written!")
storeFile.close()


print("...done!")
